from typing import Dict, List, Optional, Tuple, TypeVar, Union, cast
import sys
from typing_extensions import Literal
from zipfile import ZipFile
from autograder.config_manager import DEFAULT_ARGLIST_VALUE_KEY, ArgList, GradingConfig
from autograder.autograder import AutograderPaths
from itertools import chain
from tomlkit.api import array, document, dumps, inline_table, parse, table
from tomlkit.items import Array, Bool, InlineTable, Item, Table, item
from tomlkit.toml_document import TOMLDocument
from pathlib import Path
import shutil
from tempfile import TemporaryDirectory


def load_homeworks(dir_with_homeworks: Path):
    dir_with_homeworks = skip_inner_dirs(dir_with_homeworks)
    return [{"path": f, "name": f.name, "enabled": True} for f in dir_with_homeworks.iterdir() if f.is_file()]


def skip_inner_dirs(dir_: Path) -> Path:
    contents = list(dir_.iterdir())
    if len(contents) == 1 and contents[0].is_dir():
        return skip_inner_dirs(contents[0])
    return dir_


def extract_zip_into_dir(path, extraction_path):
    zipfile = ZipFile(path)
    zipfile.extractall(extraction_path)


def load_assignment(dir_with_assignment: str = ""):
    paths = AutograderPaths(Path(dir_with_assignment))
    if paths.tests_dir.exists() and not (
        (paths.config.exists() and paths.output_dir.exists()) or (paths.testcases_dir.exists())
    ):
        paths = AutograderPaths(paths.tests_dir)
    if not paths.config.exists() or not dir_with_assignment:
        config_file = paths.default_config
    else:
        config_file = paths.config
    grading_config = GradingConfig(config_file, paths.default_config)
    default_config = parse(paths.default_config.read_text())
    key_order = get_key_order(default_config)
    config: TOMLDocument = grading_config.file  # type: ignore
    sections = read_config_sections(config, default_config)
    sections.sort(key=lambda v: key_order[v["original_name"]])

    outputs = get_io_objects(paths.output_dir)
    inputs = get_io_objects(paths.input_dir)
    testcases = [
        {
            "name": test.name,
            "original_name": test.name,
            "text": test.read_text(),
            "input": get_single_io(test, inputs, paths.input_dir),
            "output": get_single_io(test, outputs, paths.output_dir),
            "config": read_config_sections(config, default_config, test),
        }
        # TODO: Add support for parsing stdout-only testcases
        for test in (paths.testcases_dir.iterdir() if paths.testcases_dir.exists() else [])
    ]
    return {"global_config": sections, "testcases": testcases}


T = TypeVar("T")


def get_key_order(doc: TOMLDocument) -> Dict[str, int]:
    return {v: i for i, v in enumerate(chain.from_iterable(doc[sec] for sec in doc))}  # type: ignore


def get_io_objects(dir_: Path) -> Dict[str, Dict[str, str]]:
    return {p.stem: {"path": str(p), "text": p.read_text()} for p in dir_.iterdir()} if dir_.exists() else {}


def get_single_io(test: Path, io_dict: Dict[str, Dict[str, str]], dir_: Path) -> str:
    return io_dict.get(test.stem, {})["text"] if test.stem in io_dict else ""


def read_config_sections(doc: TOMLDocument, default_doc: TOMLDocument, test: Optional[Path] = None):
    return list(chain.from_iterable(read_config_section(sec, doc, default_doc, test) for sec in doc))


def read_config_section(section_name: str, config: TOMLDocument, default_config: TOMLDocument, test=None):
    section: Table = config[section_name]  # type: ignore
    default_section: Table = default_config[section_name]  # type: ignore
    formatted_section = (
        {
            "key": prettify(k),
            "original_name": str(k),
            "value": get_value(section.value.item(k), test),
            "description": get_comment(default_section.value.item(k)),
            "section": section_name,
            **get_type(default_section.value.item(k)),
        }
        for k in section.value
    )
    if test:
        return [c for c in formatted_section if c["is_per_testcase"]]
    else:
        return list(formatted_section)


def prettify(k: str):
    return k.replace("_", " ").lower().capitalize()


def get_type(item: Item) -> dict:
    # It's a dict
    if isinstance(item, InlineTable):
        # There's a DEFAULT value
        if item:
            v = item[DEFAULT_ARGLIST_VALUE_KEY]
            if isinstance(v, str):
                type_ = "text"
            elif isinstance(v, (int, float)):
                type_ = "number"
            else:
                raise NotImplementedError(f"Field with value {item} is not supported.")
        else:
            type_ = "text"
    elif isinstance(item, Bool):
        type_ = "checkbox"
    elif isinstance(item, (str, Array)):
        type_ = "text"
    elif isinstance(item, (int, float)):
        type_ = "number"
    else:
        raise NotImplementedError(f"Field with value {item} is not supported.")
    return {"type": type_, "is_per_testcase": isinstance(item, InlineTable), "is_list": isinstance(item, Array)}


def get_value(sec, test: Optional[Path] = None):
    if isinstance(sec, Bool):
        # Tomlkit hates bools
        return bool(sec)
    elif isinstance(sec, InlineTable):
        if DEFAULT_ARGLIST_VALUE_KEY in sec and not test:
            return sec["DEFAULT"]
        elif test and test.name in sec:
            return sec[test.name]
        # This check MIGHT be unnecessary
        elif test and test.stem in sec:
            return sec[test.stem]
        else:
            return ""
    else:
        return sec


def get_comment(v: Item) -> str:
    return v.trivia.comment.lstrip("# ")


def generate_assignment_configuration(assignment, paths: AutograderPaths):
    config = document()
    paths.tests_dir.mkdir()
    paths.input_dir.mkdir()
    paths.output_dir.mkdir()
    paths.extra_dir.mkdir()
    paths.testcases_dir.mkdir()
    paths.stdout_formatters.write_text(paths.default_stdout_formatters.read_text())
    for t in assignment["testcases"]:
        name = str(t["name"])
        (paths.testcases_dir / name).write_text(t["text"])
        (paths.input_dir / name).with_suffix(".txt").write_text(t["input"])
        (paths.output_dir / name).with_suffix(".txt").write_text(t["output"])
        for c in t["config"]:
            add_value_to_config(config, c, name)
    for c in assignment["global_config"]:
        add_value_to_config(config, c)
    paths.config.write_text(dumps(config))


def build_assignment_zip(assignment, output_fname: Path):
    with TemporaryDirectory() as tmp:
        paths = AutograderPaths(Path(tmp))
        generate_assignment_configuration(assignment, paths)
        make_archive(paths.tests_dir, output_fname)


def make_archive(source: Path, destination: Path, fmt="zip"):
    if destination.exists():
        if destination.is_file():
            destination.unlink()
        else:
            raise ValueError("You can't replace a dir with a zip, duh")
    shutil.make_archive(destination.stem, fmt, source.parents[0], source.name)
    shutil.move(f"{destination.stem}.{fmt}", destination)


def add_value_to_config(config: TOMLDocument, c: dict, testcase_name: str = DEFAULT_ARGLIST_VALUE_KEY):
    value = format_config_value(c["value"], c["type"])
    if c["section"] not in config:
        config.add(c["section"], table())
    if c["original_name"] not in config[c["section"]]:
        if c["is_per_testcase"]:
            config[c["section"]].add(c["original_name"], inline_table())
        elif c["is_list"]:
            config[c["section"]].add(c["original_name"], array())
        else:
            config[c["section"]].add(c["original_name"], item(""))
    if value is None:
        return
    if c["is_per_testcase"]:
        config[c["section"]][c["original_name"]][testcase_name] = value
    else:
        config[c["section"]][c["original_name"]] = value


def format_config_value(val, typ, is_list=False):
    if is_list:
        return array([format_config_value(v, typ) for v in val.split(",") if v]) if val else array()
    else:
        if typ == "number" and val:
            return float(val)
        elif typ == "text" and val:
            return val
        elif typ == "checkbox":
            return bool(val)



import json
import os
import shutil
from typing import Dict, List, Optional, Union, overload
from autograder.autograder import AutograderPaths
import eel
from pathlib import Path

import tkinter as tk
from tkinter import filedialog
import filetype
import csv


from tempfile import TemporaryDirectory
from contextlib import redirect_stdout
from io import StringIO


from autograder_gui.util import (
    extract_zip_into_dir,
    generate_assignment_configuration,
    load_assignment,
    build_assignment_zip,
    load_homeworks,
    make_archive,
    skip_inner_dirs,
)

from autograder.__main__ import main as autograder


TEMPDIR: Optional[TemporaryDirectory] = None
CURRENT_ASSIGNMENT: Optional[dict] = None
HOMEWORK_ROOT_DIR: Optional[TemporaryDirectory] = None
HOMEWORKS: Optional[List[Dict[str, Union[str, bool]]]] = None
GRADING_RESULTS = None
PLAGIARISM_RESULTS = None

AUTOGRADER_RUN_IN_PROGRESS = "PROGRESS"

SIZE = (1200, 900)


def run():
    static_dir = (Path(__file__).parent / "static").resolve()
    eel.init(str(static_dir), allowed_extensions=[".html"])
    eel.start(
        "templates/index.html",
        jinja_templates="templates",
        size=SIZE,
        block=True,
    )


@eel.expose
def autograder_run():
    global GRADING_RESULTS
    if not HOMEWORKS or not HOMEWORK_ROOT_DIR:
        return
    if GRADING_RESULTS:
        return GRADING_RESULTS
    # Temporary value to signall that we started grading
    GRADING_RESULTS = AUTOGRADER_RUN_IN_PROGRESS

    root_dir = skip_inner_dirs(Path(HOMEWORK_ROOT_DIR.name))
    argv = ["run", str(root_dir), "-j", "-s", *(h["name"] for h in HOMEWORKS if h["enabled"])]
    generate_assignment_configuration(CURRENT_ASSIGNMENT, AutograderPaths(root_dir))
    with StringIO() as buf:
        with redirect_stdout(buf):
            try:
                autograder(argv)
            except Exception as e:
                GRADING_RESULTS = None
                return {"error": str(e)}
        GRADING_RESULTS = json.loads(buf.getvalue())
        return GRADING_RESULTS


@eel.expose
def export_grading_results():
    if GRADING_RESULTS == AUTOGRADER_RUN_IN_PROGRESS or not GRADING_RESULTS or HOMEWORK_ROOT_DIR is None:
        return

    spawn_tkinter_window()
    dst = Path(filedialog.asksaveasfilename(filetypes=[("Zip File", ".zip")]))
    root_dir = skip_inner_dirs(Path(HOMEWORK_ROOT_DIR.name))
    paths = AutograderPaths(root_dir)
    with TemporaryDirectory() as tmp:
        tmp = Path(tmp)
        if paths.results_dir.exists():
            shutil.copytree(paths.results_dir, tmp / paths.results_dir.name)
        (tmp / "results.json").write_text(json.dumps(GRADING_RESULTS, indent=4))

        with (tmp / "results.csv").open("w", newline="") as csvfile:
            spamwriter = csv.writer(csvfile, delimiter=";", quotechar="|", quoting=csv.QUOTE_MINIMAL)
            spamwriter.writerow(["submission_name", "grade"])
            for r in GRADING_RESULTS["submissions"]:
                spamwriter.writerow([r["submission"], r["final_grade"]])
        if dst.is_file():
            dst.unlink()
        make_archive(tmp, dst)


@eel.expose
def erase_grading_results():
    global GRADING_RESULTS
    if GRADING_RESULTS == AUTOGRADER_RUN_IN_PROGRESS:
        return
    if HOMEWORK_ROOT_DIR is not None:
        root_dir = skip_inner_dirs(Path(HOMEWORK_ROOT_DIR.name))
        paths = AutograderPaths(root_dir)
        shutil.rmtree(paths.tests_dir, ignore_errors=True)
    GRADING_RESULTS = None


@eel.expose
def autograder_plagiarism():
    global PLAGIARISM_RESULTS
    if not HOMEWORK_ROOT_DIR:
        return
    if not HOMEWORKS or not HOMEWORK_ROOT_DIR:
        return
    if PLAGIARISM_RESULTS:
        return PLAGIARISM_RESULTS
    # Temporary value to signal that we started checking
    PLAGIARISM_RESULTS = AUTOGRADER_RUN_IN_PROGRESS

    root_dir = skip_inner_dirs(Path(HOMEWORK_ROOT_DIR.name))
    argv = ["plagiarism", str(root_dir), "-s", *(h["name"] for h in HOMEWORKS if h["enabled"])]
    with StringIO() as buf:
        with redirect_stdout(buf):
            try:
                autograder(argv)
            except Exception as e:
                PLAGIARISM_RESULTS = None
                return {"error": str(e)}
        PLAGIARISM_RESULTS = json.loads(buf.getvalue())
        return PLAGIARISM_RESULTS


@eel.expose
def erase_plagiarism_results():
    global PLAGIARISM_RESULTS
    if PLAGIARISM_RESULTS == AUTOGRADER_RUN_IN_PROGRESS:
        return
    PLAGIARISM_RESULTS = None


@eel.expose
def export_plagiarism_results():
    if PLAGIARISM_RESULTS == AUTOGRADER_RUN_IN_PROGRESS or not PLAGIARISM_RESULTS:
        return

    spawn_tkinter_window()
    dst = Path(filedialog.asksaveasfilename(filetypes=[("Zip File", ".zip")]))
    with TemporaryDirectory() as tmp:
        tmp = Path(tmp)
        (tmp / "plagiarism_results.json").write_text(json.dumps(PLAGIARISM_RESULTS, indent=4))

        with (tmp / "plagiarism_results.csv").open("w", newline="") as csvfile:
            spamwriter = csv.writer(csvfile, delimiter=";", quotechar="|", quoting=csv.QUOTE_MINIMAL)
            spamwriter.writerow(["Submission 1", "Submission 2", "Similarity Score"])
            for r in PLAGIARISM_RESULTS["results"]:
                spamwriter.writerow([r["student1"], r["student2"], r["similarity_score"]])
        if dst.is_file():
            dst.unlink()
        make_archive(tmp, dst)


@eel.expose
def save_assignment(assignment: dict):
    global CURRENT_ASSIGNMENT
    CURRENT_ASSIGNMENT = assignment


@eel.expose
def save_homeworks(homeworks: list):
    global HOMEWORKS
    HOMEWORKS = homeworks


@eel.expose
def get_assignment():
    return CURRENT_ASSIGNMENT


@eel.expose
def get_homeworks():
    return HOMEWORKS


def spawn_tkinter_window():
    root = tk.Tk()
    root.withdraw()
    root.wm_attributes("-topmost", 1)
    return root


@eel.expose
def extract_assignment():
    global CURRENT_ASSIGNMENT

    spawn_tkinter_window().title("Choose files to open")
    chosen_paths: str = filedialog.askopenfilename(filetypes=[("Assignment file", "*.zip")])
    if not chosen_paths:
        return {"error": "No path was chosen"}
    path = Path(chosen_paths)
    if not path.is_file():
        return {"error": f"The path '{path}' is not a file."}
    with TemporaryDirectory() as tmp:
        paths = AutograderPaths(Path(tmp))
        extract_zip_into_dir(path, paths.tests_dir)
        CURRENT_ASSIGNMENT = load_assignment(tmp)
        return CURRENT_ASSIGNMENT


@eel.expose
def extract_homeworks():
    global HOMEWORKS
    global HOMEWORK_ROOT_DIR

    spawn_tkinter_window().title("Choose files to open")
    chosen_paths = filedialog.askopenfilenames(
        filetypes=[("Archive with student submissions", "*.zip"), ("Student Submissions", "*")]
    )
    if not chosen_paths:
        return {"error": "No path was chosen"}
    non_existent_paths = [f"'{p}'" for p in chosen_paths if not os.path.isfile(p)]
    if non_existent_paths:
        return {"error": f"Picked paths do not exist. Paths: {', '.join(non_existent_paths)}"}
    tmp = TemporaryDirectory()
    extraction_dir = Path(tmp.name)
    for f in chosen_paths:
        type_ = filetype.guess(f)
        if type_ and type_.extension == "zip":
            extract_zip_into_dir(chosen_paths[0], extraction_dir)
        else:
            shutil.copy(f, extraction_dir)
    HOMEWORKS = load_homeworks(extraction_dir)
    if HOMEWORK_ROOT_DIR:
        HOMEWORK_ROOT_DIR.cleanup()
    HOMEWORK_ROOT_DIR = tmp
    return HOMEWORKS


@eel.expose
def export_assignment():
    spawn_tkinter_window()
    fname = filedialog.asksaveasfilename()
    if not fname:
        return {"error": "No path was chosen"}

    build_assignment_zip(CURRENT_ASSIGNMENT, Path(fname))


@eel.expose
def create_assignment():
    global CURRENT_ASSIGNMENT
    CURRENT_ASSIGNMENT = load_assignment()
    return True


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autograder_gui']

package_data = \
{'': ['*'],
 'autograder_gui': ['static/css/*',
                    'static/images/*',
                    'static/js/*',
                    'static/lib/codemirror/css/*',
                    'static/lib/codemirror/js/*',
                    'static/lib/codemirror/mode/*',
                    'static/lib/css/*',
                    'static/lib/fontawesome/*',
                    'static/lib/fontawesome/css/*',
                    'static/lib/fontawesome/js/*',
                    'static/lib/fontawesome/less/*',
                    'static/lib/fontawesome/metadata/*',
                    'static/lib/fontawesome/scss/*',
                    'static/lib/fontawesome/sprites/*',
                    'static/lib/fontawesome/svgs/brands/*',
                    'static/lib/fontawesome/svgs/regular/*',
                    'static/lib/fontawesome/svgs/solid/*',
                    'static/lib/fontawesome/webfonts/*',
                    'static/lib/fonts/*',
                    'static/lib/fonts/static/*',
                    'static/lib/js/*',
                    'static/templates/*']}

install_requires = \
['Eel[jinja2]>=0.14.0,<0.15.0',
 'assignment-autograder>=3.0.1,<4.0.0',
 'filetype>=1.0.8,<2.0.0',
 'typing-extensions>=3.10.0,<4.0.0']

entry_points = \
{'console_scripts': ['autograder_gui = autograder_gui.__main__:run']}

setup_kwargs = {
    'name': 'autograder-gui',
    'version': '0.6.2',
    'description': '',
    'long_description': None,
    'author': 'Ovsyanka',
    'author_email': 'szmiev2000@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Ovsyanka83/autograder_electron',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
